#include "cublas.h"
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <cuda.h>
#include <sys/time.h>
#include <math.h>
#include <assert.h>

#define CHK_CUBLAS(e) { assert((e) == CUBLAS_STATUS_SUCCESS);}

void cpu_sgemm(const float *a, const float *b, 
               const int n, float *c) {

    /* this, of course, is a  
        terrible implementation */
	int i, j, k;
	for (i=0;i<n;i++) {
        for (j=0;j<n;j++) {
            double sum=0.;
            for (k=0;k<n;k++) {
                sum += a[k*n + j]*b[i*n + k];
            }
            c[i*n + j] = sum;
		}
	}
	return;
}

int get_options(int argc, char **argv, int *n, long int *seed);
void tick(struct timeval *timer);
double tock(struct timeval *timer);
void printmatrix(const char *hdr, const float *a, const int n);

int main(int argc, char **argv) {
	int n=160;
	int err;
	float *a, *b, *c, *ccuda, *ccuda2;
	float *ad, *bd, *cd;
    float *cdiff, *cdiff2;
	double abserr;
	dim3 blocksize, gridsize;
    long int seed;
    int i;
    struct timeval cputimer;
    struct timeval gputimer;
    double cputime, gputime;
    double totflops;

	err = get_options(argc, argv, &n, &seed);
	if (err) return 0;
    
    srand48(seed);

	a = (float *)malloc(n*n*sizeof(float));
	if (!a) return 1;
		
	b = (float *)malloc(n*n*sizeof(float));
	if (!b) {free(a); return 1;}

	c = (float *)malloc(n*n*sizeof(float));
	if (!c) {free(a); free(b); return 1;}

	ccuda = (float *)malloc(n*n*sizeof(float));
	if (!ccuda) {free(c); free(b); free(a); return 1;}

	cdiff = (float *)malloc(n*n*sizeof(float));
	if (!cdiff) {free(ccuda); free(c); free(b); free(a); return 1;}

	ccuda2 = (float *)malloc(n*n*sizeof(float));
	if (!ccuda2) {free(cdiff); free(ccuda); free(c); free(b); free(a); return 1;}

	cdiff2 = (float *)malloc(n*n*sizeof(float));
	if (!cdiff2) {free(ccuda2); free(cdiff); free(ccuda); free(c); free(b); free(a); return 1;}

    for (i=0; i < n*n; i++) {
        a[i] = (float)drand48();
        b[i] = (float)drand48();
    }

    totflops = 2.0*n*n*n;

	/* run CPU code */
    /* no iterations here -- calculation is big enough don't need it */

    tick(&cputimer);
    cpu_sgemm(a, b, n, c);
    cputime = tock(&cputimer);

	/* run GPU code #1 */
    cublasInit();
    CHK_CUBLAS( cublasAlloc(n*n, sizeof(float), (void**)&ad) );
    cublasAlloc(n*n, sizeof(float), (void**)&bd);
    cublasAlloc(n*n, sizeof(float), (void**)&cd);

    tick(&gputimer);

    CHK_CUBLAS( cublasSetMatrix(n, n, sizeof(float), 
                                a, n, ad, n) );
    CHK_CUBLAS( cublasSetMatrix(n, n, sizeof(float), 
                                b, n, bd, n) );
    cublasSgemm ('n', 'n', n, n, n, 1.0, ad, n, bd, n, 0.0, cd, n);
    CHK_CUBLAS( cublasGetError() );

    CHK_CUBLAS( cublasGetMatrix (n, n, sizeof(float), 
                                 cd, n, ccuda, n) );
    gputime = tock(&gputimer);

    CHK_CUBLAS( cublasFree( ad ) );
    CHK_CUBLAS( cublasFree( bd ) );
    CHK_CUBLAS( cublasFree( cd ) );
    cublasShutdown(); 

    abserr = 0.;
    for (i=0;i<n*n;i++) {
        cdiff[i] = fabs(ccuda[i] - c[i]);
        abserr += cdiff[i];
    } 
    
    if (n<20) printmatrix("Difference", cdiff, n);
    
    printf("Matrix size = %d.\n", n);
    printf("CPU  time = %lg millisec, GFLOPS=%lf\n", cputime*1000., totflops/(cputime*1.e9));
    printf("GPU  time = %lg millisec, GFLOPS=%lf, diff = %lf.\n", gputime*1000.,totflops/(gputime*1.e9),abserr);

    free(a);
    free(b);
    free(c);
    free(ccuda);
    free(cdiff);
	return 0;
}


int get_options(int argc, char **argv, int *n, long int *seed) {

  const struct option long_options[] = {
    {"matsize" , required_argument, 0, 'n'},
    {"nblocks" , required_argument, 0, 'B'},
    {"seed"    , required_argument, 0, 's'},
    {"help",      no_argument, 0, 'h'},
    {0, 0, 0, 0}};

  char c;
  int option_index;
  int tempint;

  while (1) {
    c = getopt_long(argc, argv, "n:s:h", long_options, &option_index);
    if (c == -1) break;
 
    switch(c) {
      case 'n': tempint = atoi(optarg);
          if (tempint < 1 || tempint > 50000) {
            fprintf(stderr,"%s: Cannot use number of columns %s;\n  Using %d\n", argv[0], optarg, *n);
          } else {
            *n = tempint;
          }
          break;

      case 's': *seed = atoi(optarg);
          break;

      case 'h':
          puts("Calculates y[i] = a*x[i] + b on the GPU.");
          puts("Options: ");
          puts("    --matsize=N    (-n N): Set the number of values in y,x.");
          puts("    --seed=N       (-s N): Seed the random number generator.");
          puts("");
          return +1;
        }
    }

    return 0;
}

void tick(struct timeval *timer) {
    gettimeofday(timer, NULL);
}

double tock(struct timeval *timer) {
    struct timeval now;
    gettimeofday(&now, NULL);
    return (now.tv_usec-timer->tv_usec)/1.0e6 + (now.tv_sec - timer->tv_sec);
}

void printmatrix(const char *hdr, const float *a, const int n) {
    int i,j;
    printf("%s\n", hdr);

    for (i=0; i<n; i++) {
        for (j=0; j<n; j++) {
            printf("%7.5g ", a[i*n+j]);
        } 
        puts("");
    }
}
