#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <cuda.h>
#include <sys/time.h>
#include <math.h>

#define CHK_CUDA(e) {if (e != cudaSuccess) {fprintf(stderr,"Error: %s\n", cudaGetErrorString(e)); exit(-1);}}

const int fixedblocksize=16;

void sin_n_cpu(float *c, const int nsines, const int n, const float *a) {
    int i, j;
    for (i=0; i<n; i++) {
        float val=a[i];
        for (j=0; j<nsines; j++) {
            val=sin(val); 
        }
        c[i] = val;
    }
}

__global__ 
void sin_n_fixedshared(float *cd, const int nsines, const int n, float *ad) {
    __shared__ float locdata[fixedblocksize];

    int i=threadIdx.x + blockIdx.x*blockDim.x;
    int tid=threadIdx.x;
    int j;
    
    if (i<n) {
        locdata[tid] = ad[i];
    }
    __syncthreads();
    
    if (i<n) {
        for (j=0;j<nsines;j++) {
            locdata[tid] = sin(locdata[tid]);
        }
    }
    __syncthreads();
     
    if (i<n)
        cd[i] = locdata[tid];  
}

__global__ 
void sin_n_externshared(float *cd, const int nsines, const int n, float *ad) {
    extern __shared__ float shared_data[];
    float *locdata=shared_data;

    int i=threadIdx.x + blockIdx.x*blockDim.x;
    int tid=threadIdx.x;
    int j;
    
    if (i<n) {
        locdata[tid] = ad[i];
    }
    __syncthreads();
    
    if (i<n) {
        for (j=0;j<nsines;j++) {
            locdata[tid] = sin(locdata[tid]);
        }
    }
    __syncthreads();
     
    if (i<n)
        cd[i] = locdata[tid];  

}

int get_options(int argc, char **argv, int *n, int *N, long int *seed);
void tick(struct timeval *timer);
double tock(struct timeval *timer);
void printmatrix(const char *hdr, const float *a, const int n);

int main(int argc, char **argv) {
	int n=1024;
	int N=20;
	int nblocks=10;
	int err;
	float *a, *c, *ccuda, *ccuda2;
	float *ad, *cd;
	double abserr, abserr2;
	dim3 blocksize, gridsize;
    long int seed;
    int i;
    struct timeval cputimer;
    struct timeval gputimer;
    struct timeval gputimer2;
    double cputime, gputime, gputime2;
    double totflops;

	err = get_options(argc, argv, &n, &N, &seed);
	if (err) return 0;
    
    srand48(seed);

	a = (float *)malloc(n*sizeof(float));
	if (!a) return 1;
		
	c = (float *)malloc(n*sizeof(float));
	if (!c) {free(a); return 1;}

	ccuda = (float *)malloc(n*sizeof(float));
	if (!ccuda) {free(c); free(a); return 1;}

	ccuda2 = (float *)malloc(n*sizeof(float));
	if (!ccuda2) {free(ccuda); free(c); free(a); return 1;}

    for (i=0; i < n; i++) {
        a[i] = (float)drand48();
    }

    totflops = 2.0*N*n;

	/* run CPU code */
    /* no iterations here -- calculation is big enough don't need it */
    tick(&cputimer);
    sin_n_cpu(c, N, n, a);
    cputime = tock(&cputimer);

	/* run GPU code #1 */
    CHK_CUDA( cudaMalloc(&ad, n*sizeof(float)) );
    CHK_CUDA( cudaMalloc(&cd, n*sizeof(float)) );

    tick(&gputimer);
    CHK_CUDA( cudaMemcpy(ad, a, n*sizeof(float), cudaMemcpyHostToDevice) );

    blocksize = make_uint3( fixedblocksize, 1, 1);
    nblocks = (n + fixedblocksize-1)/fixedblocksize;
    gridsize  = make_uint3( nblocks, 1, 1);

    sin_n_fixedshared<<<gridsize, blocksize>>>(cd, N, n, ad);

    CHK_CUDA( cudaMemcpy(ccuda, cd, n*sizeof(float), cudaMemcpyDeviceToHost) );
    gputime = tock(&gputimer);
        
    CHK_CUDA( cudaFree(ad) );
    CHK_CUDA( cudaFree(cd) );

	/* run GPU code #2 */
    CHK_CUDA( cudaMalloc(&ad, n*sizeof(float)) );
    CHK_CUDA( cudaMalloc(&cd, n*sizeof(float)) );

    tick(&gputimer2);
    CHK_CUDA( cudaMemcpy(ad, a, n*sizeof(float), cudaMemcpyHostToDevice) );

    blocksize = make_uint3( fixedblocksize, 1, 1);
    nblocks = (n + fixedblocksize-1)/fixedblocksize;
    gridsize  = make_uint3( nblocks, 1, 1);

    sin_n_externshared<<<gridsize, blocksize, blocksize.x*sizeof(float)>>>(cd, N, n, ad);

    CHK_CUDA( cudaMemcpy(ccuda2, cd, n*sizeof(float), cudaMemcpyDeviceToHost) );
    gputime2 = tock(&gputimer2);
        
    CHK_CUDA( cudaFree(ad) );
    CHK_CUDA( cudaFree(cd) );

    abserr = 0.;
    abserr2 = 0.;
    for (i=0;i<n;i++) {
        abserr += fabs(ccuda[i] - c[i]);
        abserr2+= fabs(ccuda2[i] - c[i]);
    }
    
    printf("Number of values = %d, number of sines = %d.\n", n, N);
    printf("CPU  time = %lg millisec, GFLOPS=%lf\n", cputime*1000., totflops/(cputime*1.e9));
    printf("GPU  time = %lg millisec, GFLOPS=%lf, diff = %lf.\n", gputime*1000.,totflops/(gputime*1.e9),abserr);
    printf("GPU2 time = %lg millisec, GFLOPS=%lf, diff = %lf.\n", gputime2*1000.,totflops/(gputime2*1.e9),abserr2);

    free(a);
    free(c);
    free(ccuda);
    free(ccuda2);
	return 0;
}


int get_options(int argc, char **argv, int *n, int *N, long int *seed) {

  const struct option long_options[] = {
    {"nvals" , required_argument, 0, 'n'},
    {"nsines" , required_argument, 0, 'N'},
    {"seed"    , required_argument, 0, 's'},
    {"help",      no_argument, 0, 'h'},
    {0, 0, 0, 0}};

  char c;
  int option_index;
  int tempint;

  while (1) {
    c = getopt_long(argc, argv, "n:s:h", long_options, &option_index);
    if (c == -1) break;
 
    switch(c) {
      case 'n': tempint = atoi(optarg);
          if (tempint < 1 || tempint > 50000) {
            fprintf(stderr,"%s: Cannot use number of values %s;\n  Using %d\n", argv[0], optarg, *n);
          } else {
            *n = tempint;
          }
          break;

      case 'N': tempint = atoi(optarg);
          if (tempint < 1 || tempint > 50000) {
            fprintf(stderr,"%s: Cannot use number of values %s;\n  Using %d\n", argv[0], optarg, *n);
          } else {
            *N = tempint;
          }
          break;

      case 's': *seed = atoi(optarg);
          break;

      case 'h':
          puts("Calculates y[i] = sin^N(x[i]) on the GPU.");
          puts("Options: ");
          puts("    --nvals=N    (-n N): Set the number of values in y,x.");
          puts("    --nsines=N   (-N N): Set the iterations of sines to use.");
          puts("    --nblocks=N  (-B N): Set the number of blocks used in each dimension.");
          puts("    --seed=N     (-s N): Seed the random number generator.");
          puts("");
          return +1;
        }
    }

    return 0;
}

void tick(struct timeval *timer) {
    gettimeofday(timer, NULL);
}

double tock(struct timeval *timer) {
    struct timeval now;
    gettimeofday(&now, NULL);
    return (now.tv_usec-timer->tv_usec)/1.0e6 + (now.tv_sec - timer->tv_sec);
}
