#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <math.h>
#include <getopt.h>

#include <CL/cl.h>

const char *kernelsrc[] = {
"__kernel void opencl_saxpb(__global const float *x,",
"	  const float a, const float b,",
"	  __global float *y)",
" {	",
"     int i = get_global_id(0);",
"     if (i < get_global_size(0) )",
"         y[i] = a*x[i] + b;",
" }"
};
const int nlines=8;

void cpu_saxpb(const float *x, float a, float b, float *y, int n) {

	int i;
	for (i=0;i<n;i++) {
		y[i] = a*x[i]+b;
	}
	return;
}

void chk(int err, char *str) {
    if (err != CL_SUCCESS) {    
        fprintf(stderr,"Error in %s\n", str); 
    }
}

void buildchk(int err, char *str) {
    switch (err) {
        case CL_SUCCESS: break;
        case CL_INVALID_PROGRAM: fprintf(stderr,"Invalid program in %s\n", str); break;
        case CL_INVALID_VALUE:   fprintf(stderr,"Invalid device value %s\n", str); break;
        case CL_INVALID_DEVICE:  fprintf(stderr,"Invalid device %s\n", str); break;
        case CL_INVALID_BINARY:  fprintf(stderr,"Invalid binary %s\n", str); break;
        case CL_INVALID_BUILD_OPTIONS: fprintf(stderr,"Invalid build options %s\n", str); break;
        case CL_INVALID_OPERATION:     fprintf(stderr,"Invalid operation %s\n", str); break;
        case CL_COMPILER_NOT_AVAILABLE: fprintf(stderr, "Compiler not available %s\n", str); break;
        case CL_BUILD_PROGRAM_FAILURE:  fprintf(stderr, "Build program failure %s\n", str); break;
        case CL_OUT_OF_HOST_MEMORY:  fprintf(stderr, "oom host %s\n", str); break;
        default: fprintf(stderr,"Unknown error in %s\n", str); break;
    }
}

int get_options(int argc, char **argv, int *n, int *nb, float *a, float *b, int *niters, int *stride);
void tick(struct timeval *timer);
double tock(struct timeval *timer);

int main(int argc, char **argv) {
	int n=1000;
	int stride=1;
	int nblocks=10;
    unsigned int blocksize;
	float a = 5.;
	float b = -1.;
	int err;
	float *x, *y, *yopencl;
	double abserr, abserr2;
    int i;
    int niters=1;
    struct timeval cputimer;
    struct timeval gputimer;
    double cputime, gputime;

	err = get_options(argc, argv, &n, &nblocks, &a, &b, &niters, &stride);
	if (err) return 0;

	blocksize = (n+nblocks-1)/nblocks;

	x = (float *)malloc(n*sizeof(float));
	if (!x) return 1;

	y = (float *)malloc(n*sizeof(float));
	if (!y) {free(x); return 1;}

	yopencl = (float *)malloc(n*sizeof(float));
	if (!yopencl) {free(y); free(x); return 1;}

    for (i=0; i<n; i++) {
        x[i] = 1.*i*i;
        yopencl[i] = 0.;
    }

	/* run CPU code */

    tick(&cputimer);
    cpu_saxpb(x, a, b, y, n);
    cputime = tock(&cputimer);

    /* create OpenCL device & context */
    cl_platform_id clPlatform;
    err = clGetPlatformIDs(1, &clPlatform, NULL);
    chk(err, "Get Platform`");

    /* query all devices available to the context */

    cl_device_id device;
    err = clGetDeviceIDs(clPlatform, CL_DEVICE_TYPE_GPU, 1, &device, NULL);
    chk(err, "Get Device IDs");

    cl_context hContext;
    hContext = clCreateContext(0, 1, &device, NULL, NULL, &err);
    chk(err, "Get Context");

    /* create a command queue for first device the context reported */
    cl_command_queue hCmdQueue;
    hCmdQueue = clCreateCommandQueue(hContext, device, 0, &err);
    chk(err, "Create Queue");

    /* create & compile program */
    cl_program hProgram;
    hProgram = clCreateProgramWithSource(hContext, nlines, kernelsrc, 0, &err);
    chk(err, "Create Program");

    err = clBuildProgram(hProgram, 1, &device, NULL, NULL, NULL);
    buildchk(err, "Build Program");

    /* create kernel */
    cl_kernel hKernel;
    hKernel = clCreateKernel(hProgram, "opencl_saxpb", &err);
    chk(err, "Create Kernel");

    /* allocate device memory */
    cl_mem yd;
    cl_mem xd;
    tick(&gputimer);
    xd = clCreateBuffer(hContext,
			CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR,
			n * sizeof(cl_float),
			x,
			&err);
    chk(err, "Create xd");

    yd = clCreateBuffer(hContext,
			CL_MEM_COPY_HOST_PTR,
			n * sizeof(cl_float),
			yopencl,
			&err);
    chk(err, "Create yd");

    /* setup parameter values */
    err  = clSetKernelArg(hKernel, 0, sizeof(cl_mem),   (void *)&xd);
    err |= clSetKernelArg(hKernel, 1, sizeof(cl_float), (void *)&a);
    err |= clSetKernelArg(hKernel, 2, sizeof(cl_float), (void *)&b);
    err |= clSetKernelArg(hKernel, 3, sizeof(cl_mem),   (void *)&yd);
    chk(err, "Set args");

    /* execute kernel */
    const size_t knsize=n;
    const size_t kblocksize=blocksize;
    err = clEnqueueNDRangeKernel(hCmdQueue, hKernel, 1, 0,
  			                     &knsize, &kblocksize, 0, 0, 0);
    chk(err, "Enqueue Kernel");

    // copy results from device back to host
    err = clEnqueueReadBuffer(hCmdQueue, yd, CL_TRUE, 0,
	  		     n * sizeof(cl_float),
			     yopencl, 0, 0, 0);
    chk(err, "Enqueue Read");

    clReleaseMemObject(xd);
    clReleaseMemObject(yd);
    clReleaseProgram(hProgram);
    clReleaseKernel(hKernel);
    clReleaseCommandQueue(hCmdQueue);
    clReleaseContext(hContext);

    gputime = tock(&gputimer);
    abserr  = 0.;
    abserr2 = 0.;
    for (i=0;i<n;i++) {
        abserr	+= fabs(yopencl[i] - y[i]);
    }

    printf("For run with n = %d, nblocks = %d, blocksize = %d, stride=%d,\n", n, nblocks, blocksize, stride);
    printf("CPU time  = %lg millisec.\n", cputime*1000.);
    printf("GPU time  = %lg millisec, diff = %lf.\n", gputime*1000., abserr);

    if (n < 20) {
        for (i=0; i<n; i++) printf("%5.3f ", y[i]); puts("");
        for (i=0; i<n; i++) printf("%5.3f ", yopencl[i]); puts("");
    }
    free(x);
    free(y);
    free(yopencl);
	return 0;
}


int get_options(int argc, char **argv, int *n, int *nb, float *a, float *b, int *niters, int *stride) {

  const struct option long_options[] = {
    {"nvals"	 , required_argument, 0, 'n'},
    {"nblocks"	 , required_argument, 0, 'B'},
    {"a", required_argument, 0, 'a'},
    {"b", required_argument, 0, 'b'},
    {"stride", required_argument, 0, 's'},
    {"help",	  no_argument, 0, 'h'},
    {0, 0, 0, 0}};

  char c;
  int option_index;
  int tempint;

  while (1) {
    c = getopt_long(argc, argv, "n:B:a:b:s:h", long_options, &option_index);
    if (c == -1) break;

    switch(c) {
      case 'n': tempint = atoi(optarg);
	  if (tempint < 1 || tempint > 500000) {
	    fprintf(stderr,"%s: Cannot use number of points %s;\n  Using %d\n", argv[0], optarg, *n);
	  } else {
	    *n = tempint;
	  }
	  break;

      case 'B': tempint = atoi(optarg);
	  if (tempint < 1 || tempint > 1000 || tempint > *n) {
	    fprintf(stderr,"%s: Cannot use number of blocks %s;\n  Using %d\n", argv[0], optarg, *nb);
	  } else {
	    *nb = tempint;
	  }
	  break;

      case 's': tempint = atoi(optarg);
	  if (tempint < 1 || tempint > *n) {
	    fprintf(stderr,"%s: Cannot use stride %s;\n	 Using %d\n", argv[0], optarg, *stride);
	  } else {
	    *stride = tempint;
	  }
	  break;

      case 'a': *a = atof(optarg);
	  break;

      case 'b': *b = atof(optarg);
	  break;

      case 'h':
	  puts("Calculates y[i] = a*x[i] + b on the GPU.");
	  puts("Options: ");
	  puts("    --nvals=N	   (-n N): Set the number of values in y,x.");
	  puts("    --nblocks=N	   (-B N): Set the number of blocks used.");
	  puts("    --a=X	   (-a X): Set the parameter a.");
	  puts("    --b=X	   (-b X): Set the parameter b.");
	  puts("    --stride=N	   (-s X): Global memory stride.");
	  puts("");
	  return +1;
	}
    }

    return 0;
}

void tick(struct timeval *timer) {
    gettimeofday(timer, NULL);
}

double tock(struct timeval *timer) {
    struct timeval now;
    gettimeofday(&now, NULL);
    return (now.tv_usec-timer->tv_usec)/1.0e6 + (now.tv_sec - timer->tv_sec);
}

