#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <cuda.h>
#include <sys/time.h>
#include <math.h>

#define CHK_CUDA(e) {if (e != cudaSuccess) {fprintf(stderr,"Error: %s\n", cudaGetErrorString(e)); exit(-1);}}

__global__ void cuda_saxpy(float *zd, const float a, float *xd, float *yd, const int n) {
}

void cpu_saxpy(float *z, const float a, const float *x, const float *y, const int n) {

	int i;
	for (i=0;i<n;i++) {
		z[i] = a*x[i]+y[i];
	}
	return;
}

int get_options(int argc, char **argv, int *n, int *nb, float *a, int *niters);
void tick(struct timeval *timer);
double tock(struct timeval *timer);

int main(int argc, char **argv) {
	int n=1000;
	int nblocks=1;
	float a = 5.;
	int err;
	float *x, *y, *z, *zcuda;
	float *xd, *yd, *zd;
	double abserr;
    int i;
    int niters=5;
    struct timeval cputimer;
    struct timeval gputimer;
    double cputime, gputime;

	err = get_options(argc, argv, &n, &nblocks, &a, &niters);
	if (err) return 0;

	x = (float *)malloc(n*sizeof(float));
	if (!x) return 1;
		
	y = (float *)malloc(n*sizeof(float));
	if (!y) {free(x); return 1;}

	z = (float *)malloc(n*sizeof(float));
	if (!z) {free(y); free(x); return 1;}

	zcuda = (float *)malloc(n*sizeof(float));
	if (!zcuda) {free(z); free(y); free(x); return 1;}

    for (i=0; i<n; i++) {
        x[i] = 1.*i;
        y[i] = 2.*i;
    }

	/* run CPU code */

    tick(&cputimer);
    for (i=0; i<niters;i++)
        cpu_saxpy(z, a, x, y, n);
    cputime = tock(&cputimer);

	/* run GPU code */
#ifdef CUDASAXPY
	CHK_CUDA( cudaMalloc(&xd, n*sizeof(float)) );
	CHK_CUDA( cudaMalloc(&yd, n*sizeof(float)) );
	CHK_CUDA( cudaMalloc(&zd, n*sizeof(float)) );

    tick(&gputimer);
	CHK_CUDA( cudaMemcpy(xd, x, n*sizeof(float), cudaMemcpyHostToDevice) );
	CHK_CUDA( cudaMemcpy(yd, y, n*sizeof(float), cudaMemcpyHostToDevice) );

    for (i=0; i<niters; i++) {
        cuda_saxpy<<<1, n>>>(zd, a, xd, yd, n);
    }

	CHK_CUDA( cudaMemcpy(zcuda, zd, n*sizeof(float), cudaMemcpyDeviceToHost) );
    gputime = tock(&gputimer);
	
	CHK_CUDA( cudaFree(xd) );
	CHK_CUDA( cudaFree(yd) );
	CHK_CUDA( cudaFree(zd) );

    abserr = 0.;
    for (i=0;i<n;i++) {
        abserr += fabs(zcuda[i] - z[i]);
    }
#endif
    
    printf("Using: n=%d, nblocks=%d, niters=%d, a=%lf\n", n, nblocks, niters, a);
    printf("CPU time = %9.4lg millisec, GFLOPS = %9.4lg\n", cputime*1000., 2.*n/(cputime*1000*1000*1000));
#ifdef CUDASAXPY
    printf("GPU time = %9.4lg millisec, GFLOPS = %9.4lg\n", gputime*1000., 2.*n/(gputime*1000*1000*1000));
    printf("CUDA and CPU results differ by %lf\n", abserr);
#endif

    free(x);
    free(y);
    free(z);
    free(zcuda);
	return 0;
}


int get_options(int argc, char **argv, int *n, int *nb, float *a, int *niters) {

  const struct option long_options[] = {
    {"nvals"     , required_argument, 0, 'n'},
    {"nblocks"   , required_argument, 0, 'B'},
    {"a", required_argument, 0, 'a'},
    {"niters", required_argument, 0, 'I'},
    {"help",      no_argument, 0, 'h'},
    {0, 0, 0, 0}};

  char c;
  int option_index;
  int tempint;

  while (1) {
    c = getopt_long(argc, argv, "n:B:a:b:h", long_options, &option_index);
    if (c == -1) break;
 
    switch(c) {
      case 'n': tempint = atoi(optarg);
          if (tempint < 1 || tempint > 500000) {
            fprintf(stderr,"%s: Cannot use number of points %s;\n  Using %d\n", argv[0], optarg, *n);
          } else {
            *n = tempint;
          }
          break;

      case 'B': tempint = atoi(optarg);
          if (tempint < 1 || tempint > 1000 || tempint > *n) {
            fprintf(stderr,"%s: Cannot use number of blocks %s;\n  Using %d\n", argv[0], optarg, *nb);
          } else {
            *nb = tempint;
          }
          break;

      case 'I': tempint = atoi(optarg);
          if (tempint < 1 || tempint > 1000 || tempint > *n) {
            fprintf(stderr,"%s: Cannot use number of iterations %s;\n  Using %d\n", argv[0], optarg, *niters);
          } else {
            *niters = tempint;
          }
          break;

      case 'a': *a = atof(optarg);
          break;

      case 'h':
          puts("Calculates z[i] = a*x[i] + y[i] on the GPU.");
          puts("Options: ");
          puts("    --nvals=N      (-n N): Set the number of values in y,x.");
          puts("    --nblocks=N    (-B N): Set the number of blocks used.");
          puts("    --a=X          (-a X): Set the parameter a.");
          puts("    --niters=N     (-I X): Set number of iterations to calculate.");
          puts("");
          return +1;
        }
    }

    return 0;
}

void tick(struct timeval *timer) {
    gettimeofday(timer, NULL);
}

double tock(struct timeval *timer) {
    struct timeval now;
    gettimeofday(&now, NULL);
    return (now.tv_usec-timer->tv_usec)/1.0e6 + (now.tv_sec - timer->tv_sec);
}
